import typing
import functools
import traceback

from ..definition.contracts import Error


def semantic(kind: typing.Literal['usecase', 'io', 'auditor', 'context', 'response', 'reader', 'writer']):
    def decorator(func):
        func.__dict__['__semantic__'] = kind

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            return await func(*args, **kwargs)

        return wrapper

    return decorator


def skip(func):
    func.__dict__['__semantic__'] = 'skip'

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        return await func(*args, **kwargs)

    return wrapper


def default_exception(e: Exception, add_traceback: bool = False) -> Error:
    return Error(
        status_code=500,
        message=str(e),
        details=''.join(traceback.format_tb(e.__traceback__)) if add_traceback else None
    )



import collections.abc

import zorge
import zodchy

from . import registry, processing


class ExpectedErrorOccurred(Exception):
    pass


class TaskExecutor:
    def __init__(
        self,
        di_container: zorge.Container,
        actor_registry: registry.ActorRegistry,
        error_wrapper: collections.abc.Callable[[Exception], zodchy.codex.cqea.Error] | None = None
    ):
        self._di_container = di_container
        self._actor_registry = actor_registry
        self._error_wrapper = error_wrapper

    async def run(
        self,
        task: zodchy.codex.cqea.Task,
        execution_context: registry.ExecutionContext | None = None
    ) -> list[zodchy.codex.cqea.Message]:
        resolver_context = (execution_context,) if execution_context else ()
        stream = []
        try:
            async with self._di_container.get_resolver(*resolver_context) as resolver:
                processor = processing.CQProcessor(self._actor_registry, resolver)(task)
                async for message in processor:
                    stream.append(message)
                    if isinstance(message, zodchy.codex.cqea.Error):
                        raise ExpectedErrorOccurred  # raise error just for informing context manager
        except ExpectedErrorOccurred:
            pass  # supress this artificial error
        except Exception as e:
            if self._error_wrapper:
                stream.append(self._error_wrapper(e))
            else:
                raise e
        return stream



import dataclasses
import typing

from zodchy import codex


@dataclasses.dataclass(frozen=True)
class Command(codex.cqea.Command):
    pass


@dataclasses.dataclass(frozen=True)
class Query(codex.cqea.Query):
    pass


@dataclasses.dataclass(frozen=True)
class Context(codex.cqea.Context):
    pass


@dataclasses.dataclass(frozen=True)
class Event(codex.cqea.Event):
    pass


@dataclasses.dataclass(frozen=True)
class Error(codex.cqea.Error):
    status_code: int
    message: str
    semantic_code: int | None = None
    details: typing.Any | None = None



from .implementation import (
    TaskExecutor,
    CQProcessor,
    ActorRegistry,
    register_module
)
from .definition.contracts import (
    Error,
    Command,
    Query,
    Event,
    Context
)
from .aux import wrappers
